/*
 * $Id: Mozilla2Pilot.c,v 1.4 2004/03/12 21:23:39 janusz Exp $
 *
 * This is source code of simple "conduit", which converts  Mozilla's 
 * offline news file into TuNews database.
 *
 * This file is licensed under Gnu Public License (GPL)
 *
 * (C) 2004 Janusz Zamecki
 *
 *
 * To compile: 
 * 	Linux: gcc -Wall -Os -s Mozilla2Pilot.c -o Mozilla2Pilot
 * 
 * BUGS: it is not bullet proof
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifndef __MINGW32__
#include <endian.h>
#else
#define __LITTLE_ENDIAN 1234
#define __BIG_ENDIAN 4321
#define __BYTE_ORDER __LITTLE_ENDIAN /* x86 */
#endif

#ifndef __BIG_ENDIAN
#define __BIG_ENDIAN 4321
#endif
#ifndef __LITTLE_ENDIAN
#define __LITTLE_ENDIAN 1234
#endif

#if __BYTE_ORDER == __BIG_ENDIAN
#define htons(A) (A)
#define htonl(A) (A)
#elif __BYTE_ORDER == __LITTLE_ENDIAN
#define htons(A)  ((((unsigned short)(A)) >> 8) | (((unsigned short)(A)) << 8))
#define htonl(A)  ((((unsigned long)(A)) >> 24) | (((unsigned long)(A) & 0x00ff0000) >> 8) | \
		  (((unsigned long)(A) & 0x0000ff00) << 8) | (((unsigned long)(A)) << 24))
#else
#error Please define your endianess
#endif


#define ART_DELIM "From - "	/* Delimiter line in Mozilla's newsgroup file */
#define OE_DELIM "From liboe@linux " /* Delimiter line in OE newsgroup file converted by oe2mbx */
#define OE_DELIM_L 17
#define ART_DELIM_L 7		/* lenght of ART_DELIM */
#define BUF_SIZE 4096		/* max length of line */

#define ART_MSGID_OFFSET 10	/* offset to msgId in article record */

void Usage(void) {
	printf("Usage: Mozilla2Pilot [-f ID] [-g group_name] offline_news_file [TuNews_database]\n");
	printf("\nExample: Mozilla2Pilot comp.os.advocacy\n");
	printf("\nIf offline_news_file is not a news group name, use -g switch.\n");
	printf("\nExample: Mozilla2Pilot -g comp.os.advocacy articles.txt\n\n");
	exit(100);
}

/* scans input file and counts articles starting from article_id 
 *
 * returns # of articles in input file starting from article with given ID
 */
unsigned long scan_input_file(FILE *f_in, char *groupname, unsigned long article_id) {
	char buf[BUF_SIZE], *ptr;
	unsigned long beg_pos=0, counter=0, grlen=strlen(groupname);

	if (fgets(buf, BUF_SIZE, f_in)) {
		if (strncmp(buf, ART_DELIM, ART_DELIM_L)!=0 && strncmp(buf, OE_DELIM, OE_DELIM_L)!=0) {
			printf("\n\nInput file is not mbx offline newsgroup file.\n");
			exit(90);
		}
	}

	while (fgets(buf, BUF_SIZE, f_in)) {
		if (strncmp(buf, ART_DELIM, ART_DELIM_L)==0 || strncmp(buf, OE_DELIM, OE_DELIM_L)==0) {
			if (counter)
				counter++;
			else
				beg_pos=ftell(f_in)-strlen(buf);
		} else if (counter==0 && strncmp(buf, "Xref:", 5)==0) {
			ptr=buf+strlen(buf)-1;
			if (*ptr=='\n') *ptr--;
			if (*ptr=='\r') *ptr--;
			while (ptr>buf && *ptr>='0' && *ptr<='9')
				ptr--;
			*ptr='\0';
			if (strcmp(ptr-grlen, groupname)==0) /* ignore other newsgroups */
				if (atol(++ptr)>article_id)
					counter++;
		}
	}
	fseek(f_in, beg_pos, SEEK_SET);
	return counter;
}

/* creates .pdb file header */
int create_output_header(FILE *f_out, char *group_name, unsigned short articles) {
	char buf[32];
	unsigned short art_cnt;
	unsigned long offset;

	/* header description: http://www.palmos.com/dev/support/docs/fileformats/PDB+PRCFormat.html */
	
	buf[0]=buf[1]=buf[2]='$';	/* required by TuNews import feature */
	strncpy(buf+3, group_name, 28);
	fwrite(buf, 32, 1, f_out);	/* offset 0:  db name */

	memset(buf, 0, 3);
	buf[3]=1;

	fwrite(buf, 4, 1, f_out);	/* offset 32: attributes & version */
	fwrite(buf, 4, 1, f_out);	/* offset 36: fake creation time */
	fwrite(buf, 4, 1, f_out);	/* offset 40: fake modyfication time */
	fwrite(buf, 4, 1, f_out);	/* offset 44: fake last backup time - we don't need this time values */

	memset(buf, 0, ART_MSGID_OFFSET);

	fwrite(buf, 4, 1, f_out);       /* offset 48: modyfication number */
	fwrite(buf, 4, 1, f_out);       /* offset 52: AppInfoArea */
	fwrite(buf, 4, 1, f_out);       /* offset 56: SortInfoArea */

	strcpy(buf+ART_MSGID_OFFSET, "Arts");
	fwrite(buf+ART_MSGID_OFFSET, 4, 1, f_out);	/* offset 60 */

	strcpy(buf+ART_MSGID_OFFSET, "NeWS");
	fwrite(buf+ART_MSGID_OFFSET, 4, 1, f_out);	/* offset 64 */

	fwrite(buf, 4, 1, f_out);       /* offset 68: Unique ID Seed */
	fwrite(buf, 4, 1, f_out);       /* offset 72: NextRecordList ID */

	articles++;
	art_cnt=htons(articles);	/* pdb is big endian */
	fwrite(&art_cnt, 2, 1, f_out);	/* offset 76: record count */

	/* first record list element */
	offset=(unsigned long)(articles)*8+78+2;	/* offset of 1st record: header length + articles list length + place holder */
	offset=htonl(offset);
	fwrite(&offset, 4, 1, f_out);
	fwrite(buf, 4, 1, f_out);
	
	while(--articles)		/* write empty record list */
		fwrite(buf, 8, 1, f_out);
	
	fwrite(buf, 2, 1, f_out);	/* place holder */
	
	/* write special record with group name */
	fwrite(buf, ART_MSGID_OFFSET, 1, f_out);				/* write reserved article header - must be filled with 0s */
	fwrite("@@TuNews@@", 11, 1, f_out);					/* write with trailing 0 */
	fwrite(group_name, strlen(group_name)+1, 1, f_out);			/* write with trailing 0 */

	return 0;
}

/* copies one article from f_in to f_out and updates pdb header 
 *
 * Returns 0 on success
 */
int copy_article(FILE *f_in, FILE *f_out, char *newsgroup, unsigned short article, unsigned long *article_id) {
	char buf[BUF_SIZE], *ptr;
	char header[2*BUF_SIZE];	/* should be enough for stripped header */
	char message_id[128];
	int copy_header_mode=1, copy_line=0, newsgroup_length=strlen(newsgroup), article_ok=0; 
	unsigned long pos;

	/* f_in position is on ART_DELIM, skip it */
	if (fgets(buf, BUF_SIZE, f_in)==NULL)
		return -1;

	/* f_out position is at the end of file */
	pos=ftell(f_out);
	pos=htonl(pos); /* this is macro! */
	fseek(f_out, 78+8*article, SEEK_SET);	/* write offset in record list */
	fwrite(&pos, 4, 1, f_out);
	fseek(f_out, 0, SEEK_END);

	memset(header, 0, ART_MSGID_OFFSET);

	fwrite(header, ART_MSGID_OFFSET, 1, f_out);

	/* TODO: checking for buffer overflow in header table */

	/* fill the header firs */
	while (fgets(buf, BUF_SIZE, f_in)) {
		/* copy needed header fields only */
		if (copy_header_mode) {
			if (strlen(buf)<5) {	/* end of header? */
				strcat(header, "\r\n");
				copy_header_mode=0;
				if (article_ok) {
					fwrite(message_id, strlen(message_id)+1, 1, f_out);
					fwrite(header, strlen(header), 1, f_out);
				}
			} else if (strncmp(buf, "Xref:", 5)==0) {
				ptr=buf+strlen(buf)-1;
				if (*ptr=='\n') *ptr--;
				if (*ptr=='\r') *ptr--;
				while (ptr>buf && *ptr>='0' && *ptr<='9')
					ptr--;
				*article_id=atol(++ptr);
				copy_line=1;
			} else if (copy_line && *buf=='\t') {
				;
			} else if (strncmp(buf, "From:", 5)==0 ||
			    strncmp(buf, "Date:", 5)==0 ||
			    strncmp(buf, "Subject:", 8)==0 ||
			    strncmp(buf, "References:", 11)==0) {
				copy_line=1;
			} else if (strncmp(buf, "Newsgroups:", 11)==0) {
				copy_line=1;
				if ((ptr=strstr(buf+11, newsgroup))!=NULL) {
					ptr+=newsgroup_length;
					if (*ptr=='\r' || *ptr=='\n' || *ptr==' ')
						article_ok=1;
					else
						article_ok=0;
				}
			} else if (strncmp(buf, "Message-ID:", 11)==0) {
				ptr=buf+strlen(buf)-1;
				if (*ptr=='\n') *ptr--='\0';
				if (*ptr=='\r') *ptr--='\0';

				while (ptr>buf && *ptr!=' ' && *ptr!=':') ptr--;
				
				strncpy(message_id, ++ptr, 127);
				strcat(header, buf);
				strcat(header, "\r\n");		/* TuNews needs it */
				
				copy_line=0;
			} else
				copy_line=0;

			if (copy_line) {
				ptr=buf+strlen(buf)-1;
				if (*ptr=='\n') *ptr--='\0';
				if (*ptr=='\r') *ptr--='\0';
			
				strcat(header, buf);
				strcat(header, "\r\n");		/* TuNews needs it */
			}
		} else if  (strncmp(buf, ART_DELIM, ART_DELIM_L)==0 || strncmp(buf, OE_DELIM, OE_DELIM_L)==0) {
			fseek(f_in, -strlen(buf), SEEK_CUR);
			*buf='\0';
			fwrite(buf, 1, 1, f_out);		/* TuNews needs ending 0 */
			return 0;
		} else {
			ptr=buf+strlen(buf)-1;
			if (*ptr=='\n') *ptr--='\0';
			if (*ptr=='\r') *ptr--='\0';
			strcat(buf, "\r\n");

			fwrite(buf, strlen(buf), 1, f_out);
		}
	}

	*buf='\0';
	fwrite(buf, 1, 1, f_out);		/* TuNews needs ending 0 */

	return 0;
}

/*
 * exit values:
 * 		0 - OK
 * 		1 - cannot open input file
 * 		2 - cannot open output file
 * 	       80 - no articles to write
 * 	       90 - input file format error
 * 	      100 - usage error
 *
 */
int main(int argc, char **argv) {
	char of_name_buf[256], *ptr;
	char group_name[128];
	FILE *f_in, *f_out;
	unsigned long article_id; 
	unsigned short articles, i;


	if (argc==1) Usage();

	/* parse command line arguments */
	if (strcmp(argv[1], "-f")==0) {
		article_id=atol(argv[2]);
		argv+=2;
		argc-=2;
	} else
		article_id=0;

	if (strcmp(argv[1], "-g")==0) {
		strncpy(group_name, argv[2], 127);
		argv+=2;
		argc-=2;
	} else { /* file name is a group name */
		for (ptr=argv[1]+strlen(argv[1])-1; ptr>argv[1] && *ptr!='/' && *ptr!='\\'; ptr--);
		if (*ptr=='/' || *ptr=='\\')
			ptr++;
	
		strncpy(group_name, ptr, 127);
	}
	
	if ((f_in=fopen(argv[1], "rb"))==NULL) {	
		printf("Cannot open input file: ");
		printf(argv[1]);
		printf("\n");
		exit(1);
	}

	printf("Input file: \t");
	printf(argv[1]);
	printf("\nGroup name: \t");
	printf(group_name);
	printf("\nOutput file: \t");

	if (--argc>1) {
		strncpy(of_name_buf, argv[2], 255-4);
		if (!strstr(of_name_buf, ".pdb"))
			strcat(of_name_buf, ".pdb");
		argv++;
	} else {
		strncpy(of_name_buf, group_name, 255-4);
		strcat(of_name_buf, ".pdb");
	}

	printf(of_name_buf);

	/* do real job now */
	articles=scan_input_file(f_in, group_name, article_id);

	printf("\n# of articles in output file: %u\n", articles);
	fflush(NULL);

	if (articles==0) {
		printf("Nothing to do.\n");
		exit(80);
	}
		
	if ((f_out=fopen(of_name_buf, "wb+"))==NULL) {
		printf("\nCannot open output file: %s\n", of_name_buf);
		exit(2);
	}

	create_output_header(f_out, group_name, articles);

	for (i=1; i<=articles; i++) {	/* record #0 is special */
		copy_article(f_in, f_out, group_name, i, &article_id);
	}

	fclose(f_in);
	fclose(f_out);

	printf("Last article ID: %lu\n\n", article_id);

	exit(0);
}
